import glob
import os
import subprocess
import time
from typing import Any, Dict, List, Optional, Tuple

import psutil
import yaml
from streamlit_autorefresh import st_autorefresh

import streamlit as st

st.set_page_config(
    page_title="Web Runner",
    page_icon=":tornado:",
    layout="centered",
    initial_sidebar_state="auto",
)


def check_process_running(pid: int) -> bool:
    try:
        proc = psutil.Process(pid)
        if proc.status() == "zombie":
            proc.terminate()
            return False
        return True
    except psutil.NoSuchProcess:
        return False


def stop_running_task(pid: int):
    proc = psutil.Process(pid)
    for child in proc.children(recursive=True):
        child.terminate()
    proc.terminate()


def get_running_pid() -> Tuple[Optional[str], Optional[int], bool]:
    if not os.path.isfile("running.pid"):
        return None, None, False
    with open("running.pid") as f:
        fields = f.read().split()
        task_id = fields[0]
        pid = int(fields[1])
    finished = not check_process_running(pid)
    return task_id, pid, finished


def get_datasets() -> List[str]:
    datasets = []
    for p in glob.glob("data/*.hdf5"):
        if "euclidean" in p or "angular" in p:
            datasets.append(os.path.basename(p)[:-5])
    return datasets


def get_template() -> Dict[str, Any]:
    # load algos.yaml template
    with open("algos_template.yaml") as f:
        raw_data = f.read()

    return yaml.safe_load(raw_data)


def get_task_id() -> str:
    import datetime

    now = datetime.datetime.now()
    return now.strftime("%Y%m%d-%H%M%S")


def get_algorithm_key(algorithm: str, float16: bool, autogc: bool) -> str:
    key = f"tairvector-{algorithm.lower()}"
    if float16:
        key += "-f16"
    if autogc:
        key += "-autogc"
    return key


def create_profile(
    filename: str,
    url: str,
    algorithm: str,
    key: str,
    index_args,
    query_args,
):
    algorithm_profile = {
        "docker-tag": "ann-benchmarks-tvs",
        "module": "ann_benchmarks.algorithms.tairvector",
        "constructor": "TairVector",
        "base-args": [
            {"url": url, "parallelism": 4},
            "@metric",
            algorithm,
        ],
        "run-groups": {"default": {"args": [index_args], "query_args": [query_args]}},
    }

    yaml.CDumper.ignore_aliases = lambda *args: True
    profile_data = yaml.dump(
        {
            "float": {
                "any": {key: algorithm_profile},
                "euclidean": {},
                "angular": {},
            }
        },
        Dumper=yaml.CDumper,
    )
    with open(filename, "w+") as f:
        f.write(profile_data)


def start_task(task_id: str, algo: str, batch: bool, datasets: List[str]):
    proc = subprocess.Popen(
        ["./run.sh", task_id, algo, "1" if batch else "0"] + datasets,
        start_new_session=True,
    )
    # write task_id & pid
    with open("running.pid", "w+") as f:
        f.write(f"{task_id} {proc.pid}")


def init_env():
    if not os.path.exists("profile"):
        os.mkdir("profile")
    elif not os.path.isdir("profile"):
        raise RuntimeError("Cannot create ./profile")
    if not os.path.exists("log"):
        os.mkdir("log")
    elif not os.path.isdir("log"):
        raise RuntimeError("Cannot create ./log")


init_env()

# read pid file & check if there is a running benchmark
task_id, pid, finished = get_running_pid()

if pid is not None:
    row1_col1 = st.columns(1)[0]
    row2_col1 = st.columns(1)[0]
    row3_col1, row3_col2, _ = st.columns([2, 2, 10])
    if finished:
        with row1_col1:
            st.header(f"Task #{task_id} Finished")
        with row3_col1:
            pass
        with row3_col2:
            if st.button("New", use_container_width=True):
                os.unlink("running.pid")
                st.experimental_rerun()
    else:
        with row1_col1:
            st.header(f"Task #{task_id} Running")
        with row3_col1:
            pass
        with row3_col2:
            if st.button("Stop", use_container_width=True):
                stop_running_task(pid)
                os.unlink("running.pid")
                st.experimental_rerun()
        st_autorefresh(5000)
    with row2_col1:
        with open(f"log/{task_id}.log") as f:
            log_data = f.read()
            st.text_area("LOG", log_data, label_visibility="hidden", height=300, disabled=True)
    st.stop()


url = st.session_state.get("tair_url", "")
url = st.text_input("Tair URL", placeholder="redis://", value=url)
st.session_state["tair_url"] = url
# parallelism = st.number_input("Parallelism", min_value=2, max_value=8)

available_datasets = get_datasets()
selected_datasets = st.multiselect("Datasets", available_datasets)

available_algoritms = ["HNSW", "FLAT"]
algorithm = st.selectbox("Algorithm", available_algoritms)

profiles = get_template()

index_args = None
query_args = None
if algorithm == "HNSW":
    default_run_group = profiles["float"]["any"][algorithm]["run-groups"]["default"]
    index_args = default_run_group["args"][0]
    query_args = default_run_group["query-args"][0]

use_float16 = st.checkbox("Use Float16")

enable_autogc = False
if algorithm == "HNSW":
    enable_autogc = st.checkbox("Enable Auto GC")

multi_thread = st.checkbox("Multi-Thread KNN Search")


selected_index_args = []
selected_query_args = []
if index_args is not None:
    selected_index_args = st.multiselect("Index Args", index_args, default=index_args, disabled=False)

if query_args is not None:
    selected_query_args = st.multiselect("Query Args", query_args, default=query_args, disabled=False)

if st.button("Start"):
    if url == "" or algorithm is None or len(selected_datasets) == 0:
        st.warning("Missing or invalid paramters")
    else:
        if use_float16:
            for args in selected_index_args:
                args["data_type"] = "FLOAT16"
        if enable_autogc:
            for args in selected_index_args:
                args["auto_gc"] = 1
        task_id = get_task_id()
        algorithm_key = get_algorithm_key(algorithm, use_float16, enable_autogc)
        create_profile(
            f"profile/algos-{task_id}.yaml",
            url,
            algorithm,
            algorithm_key,
            selected_index_args,
            selected_query_args,
        )
        start_task(
            task_id,
            algorithm_key,
            multi_thread,
            selected_datasets,
        )
        time.sleep(1)
        st.experimental_rerun()
